#!/usr/bin/env python2.7
#
# CDDL HEADER START
#
# The contents of this file are subject to the terms of the
# Common Development and Distribution License, version 1.1 (the "License").
# You may not use this file except in compliance with the License.
#
# You can obtain a copy of the license at ./LICENSE.
# See the License for the specific language governing permissions
# and limitations under the License.
#
# When distributing Covered Code, include this CDDL HEADER in each
# file and include the License file at ./LICENSE.
# If applicable, add the following below this CDDL HEADER, with the
# fields enclosed by brackets "[]" replaced with your own identifying
# information: Portions Copyright [yyyy] [name of copyright owner]
#
# CDDL HEADER END
#

# Copyright (c) 2015 Daniel W. Steinbrook. All rights reserved.

import sys
import argparse
from libzzzfs import zfs
from libzzzfs.dataset import Dataset, Pool
from libzzzfs.util import PropertyAssignment, PropertyList, ZzzFSException


def zzzfs_main(argv):
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(dest='command', title='subcommands')

    # per-command arguments
    clone = subparsers.add_parser(
        'clone', help='turn a snapshot into a filesystem with a new name')
    clone.add_argument('snapshot')
    clone.add_argument('filesystem')

    create = subparsers.add_parser('create', help='create a filesystem')
    create.add_argument('filesystem')
    create.add_argument(
        '-p', action='store_true', dest='create_parents',
        help='create missing parent filesystems')
    create.add_argument(
        '-o', metavar='property=value', action='append', dest='properties',
        default=[], type=PropertyAssignment, help='set the specified property')

    destroy = subparsers.add_parser('destroy', help='destroy a filesystem')
    destroy.add_argument('filesystem')
    destroy.add_argument(
        '-r', action='store_true', dest='recursive',
        help='destroy child filesystems')

    diff = subparsers.add_parser(
        'diff', help='compare filesystem/snapshot against a snapshot')
    diff.add_argument('identifier', metavar='snapshot')
    diff.add_argument('other_identifier', metavar='snapshot|filesystem')

    get = subparsers.add_parser('get', help='get dataset properties')
    recursive_or_depth = get.add_mutually_exclusive_group()
    recursive_or_depth.add_argument(
        '-r', action='store_true', dest='recursive',
        help='display all children')
    recursive_or_depth.add_argument(
        '-d', metavar='depth', type=int, dest='max_depth', default=0,
        help='number of child generations to display')
    get.add_argument(
        'properties', metavar='all | property[,property...]', type=PropertyList,
        help='comma-separated list of properties')
    get.add_argument(
        'identifiers', metavar='filesystem|snapshot', nargs='+')
    get.add_argument(
        '-H', action='store_true', dest='scriptable_mode',
        help='scripted mode (no headers, tab-delimited)')
    get.add_argument(
        '-o', metavar='all | field[,field...]', type=PropertyList,
        default=PropertyList('all'), dest='headers',
        help='comma-separated list of fields (name, property, value, source)')
    get.add_argument(
        '-t', metavar='type[,type...]', dest='types', type=PropertyList,
        default=PropertyList('filesystems'),
        help='comma-separated list of types (all, filesystems, snapshots)')
    get.add_argument(
        '-s', metavar='source[,source...]', type=PropertyList, dest='sources',
        default=PropertyList('local,inherited'),
        help='comma-separated list of sources (local, inherited)')

    inherit = subparsers.add_parser(
        'inherit', help='unset a property from datasets')
    inherit.add_argument('property')
    inherit.add_argument(
        'identifiers', metavar='filesystem|snapshot', nargs='+')

    list_ = subparsers.add_parser('list', help='list datasets')
    recursive_or_depth = list_.add_mutually_exclusive_group()
    recursive_or_depth.add_argument(
        '-r', action='store_true', dest='recursive',
        help='display all children')
    recursive_or_depth.add_argument(
        '-d', metavar='depth', type=int, dest='max_depth', default=0,
        help='number of child generations to display')
    list_.add_argument(
        '-H', action='store_true', dest='scriptable_mode',
        help='scripted mode (no headers, tab-delimited)')
    list_.add_argument(
        '-o', metavar='property[,property...]', dest='headers',
        type=PropertyList, help='comma-separated list of properties',
        default=PropertyList('name,used,available,refer,mountpoint'))
    list_.add_argument(
        '-t', metavar='type[,type...]', dest='types', type=PropertyList,
        default=PropertyList('filesystems'),
        help='comma-separated list of types (all, filesystems, snapshots)')
    list_.add_argument(
        '-s', metavar='property', dest='sort_asc', action='append', default=[],
        help='sort by property (ascending)')
    list_.add_argument(
        '-S', metavar='property', dest='sort_desc', action='append', default=[],
        help='sort by property (descending)')
    list_.add_argument('identifiers', metavar='filesystem|snapshot', nargs='*')

    promote = subparsers.add_parser(
        'promote', help='turn a cloned snapshot into a standalone filesystem')
    promote.add_argument('clone_filesystem')

    receive = subparsers.add_parser(
        'receive', help='create a new filesystem from "zzzfs send" output')
    receive.add_argument('filesystem')

    rename = subparsers.add_parser('rename', help='move or rename a dataset')
    rename.add_argument('identifier', metavar='filesystem|snapshot')
    rename.add_argument('other_identifier', metavar='filesystem|snapshot')

    rollback = subparsers.add_parser(
        'rollback', help='replace a filesystem with a snapshot')
    rollback.add_argument('snapshot')

    send = subparsers.add_parser(
        'send', help='serialize snapshot into a data stream')
    send.add_argument('snapshot')

    set_ = subparsers.add_parser(
        'set', help='set a property value for a dataset')
    set_.add_argument(
        'keyval', metavar='property=value', type=PropertyAssignment)
    set_.add_argument('identifiers', metavar='filesystem|snapshot', nargs='+')

    snap = subparsers.add_parser(
        'snapshot', help='create snapshots of filesystems')
    snap.add_argument('snapshots', metavar='filesystem@snapname', nargs='+')
    snap.add_argument(
        '-o', metavar='property=value', action='append', dest='properties',
        default=[], type=PropertyAssignment, help='set the specified property')

    # generate dict of argument keys/values
    args = parser.parse_args(argv[1:])
    params = dict(args._get_kwargs())
    del params['command']

    if args.command is None:
        sys.exit(parser.print_usage())

    retval = getattr(zfs, args.command)(**params)

    if type(retval) is str:
        return retval

    elif args.command not in ('diff', 'get', 'list', 'send'):
        # pool-modifying commands; log in pool history
        if isinstance(retval, Dataset):
            retval.pool.log_history_event(argv)
        else:
            # multiple affected datasets; only log command once per pool
            for pool_name in list(set(dataset.pool.name for dataset in retval)):
                Pool(pool_name).log_history_event(argv)


def main():
    try:
        output = zzzfs_main(sys.argv)
    except ZzzFSException as e:
        sys.exit('%s: %s' % (sys.argv[0], e))

    if output:
        print(output)


if __name__ == '__main__':
    main()



#!/usr/bin/env python2.7
#
# CDDL HEADER START
#
# The contents of this file are subject to the terms of the
# Common Development and Distribution License, version 1.1 (the "License").
# You may not use this file except in compliance with the License.
#
# You can obtain a copy of the license at ./LICENSE.
# See the License for the specific language governing permissions
# and limitations under the License.
#
# When distributing Covered Code, include this CDDL HEADER in each
# file and include the License file at ./LICENSE.
# If applicable, add the following below this CDDL HEADER, with the
# fields enclosed by brackets "[]" replaced with your own identifying
# information: Portions Copyright [yyyy] [name of copyright owner]
#
# CDDL HEADER END
#

# Copyright (c) 2015 Daniel W. Steinbrook. All rights reserved.

import sys
import argparse

from libzzzfs import zpool
from libzzzfs.dataset import Pool, ZzzFSException
from libzzzfs.util import PropertyList


def zzzpool_main(argv):
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(dest='command', title='subcommands')

    # per-command arguments
    create = subparsers.add_parser('create', help='create a pool')
    create.add_argument('pool_name', metavar='pool', help='pool name')
    create.add_argument('disk', help='directory in which to create pool')

    destroy = subparsers.add_parser('destroy', help='destroy a pool')
    destroy.add_argument('pool_name', metavar='pool', help='pool name')

    history = subparsers.add_parser(
        'history', help='display pool command history')
    history.add_argument(
        'pool_names', metavar='pool', nargs='*', default=[], help='pool name')
    history.add_argument(
        '-l', action='store_true', dest='long_format',
        help='show log records in long format')

    list_ = subparsers.add_parser('list', help='list pools and properties')
    list_.add_argument('pool_name', nargs='?', default=None, help='pool name')
    list_.add_argument(
        '-H', action='store_true', dest='scriptable_mode',
        help='scripted mode (no headers, tab-delimited)')
    list_.add_argument(
        '-o', metavar='property[,...]', type=PropertyList, dest='headers',
        default=PropertyList('name,size,alloc,free,cap,health,altroot'),
        help='comma-separated list of properties')

    # generate dict of argument keys/values
    args = parser.parse_args(argv[1:])
    params = dict(args._get_kwargs())
    del params['command']

    if args.command is None:
        sys.exit(parser.print_usage())

    retval = getattr(zpool, args.command)(**params)
    if type(retval) is str:
        return retval

    if isinstance(retval, Pool) and args.command == 'create':
        retval.log_history_event(argv)


def main():
    try:
        output = zzzpool_main(sys.argv)
    except ZzzFSException as e:
        sys.exit('%s: %s' % (sys.argv[0], e))

    if output:
        print(output)


if __name__ == '__main__':
    main()



#!/usr/bin/env python2.7
#
# CDDL HEADER START
#
# The contents of this file are subject to the terms of the
# Common Development and Distribution License, version 1.1 (the "License").
# You may not use this file except in compliance with the License.
#
# You can obtain a copy of the license at ./LICENSE.
# See the License for the specific language governing permissions
# and limitations under the License.
#
# When distributing Covered Code, include this CDDL HEADER in each
# file and include the License file at ./LICENSE.
# If applicable, add the following below this CDDL HEADER, with the
# fields enclosed by brackets "[]" replaced with your own identifying
# information: Portions Copyright [yyyy] [name of copyright owner]
#
# CDDL HEADER END
#

# Copyright (c) 2015 Daniel W. Steinbrook. All rights reserved.

import os
import sys
import shutil
import filecmp

from libzzzfs.dataset import (
    get_all_datasets, get_dataset_by, Filesystem, Pool, Snapshot)
from libzzzfs.util import tabulated, validate_component_name, ZzzFSException


# Each method returns a string to be written to stdout, or a dataset (or list
# of datasets) affected by the command.

def clone(snapshot, filesystem):
    '''Turn a snapshot into a filesystem with a new name.'''
    dataset1 = get_dataset_by(snapshot, should_be=Snapshot)
    dataset2 = get_dataset_by(
        filesystem, should_be=Filesystem, should_exist=False)

    dataset1.clone_to(dataset2)
    return [dataset1, dataset2]


def create(filesystem, create_parents, properties):
    '''Create a filesystem.'''
    dataset = get_dataset_by(
        filesystem, should_be=Filesystem, should_exist=False)

    dataset.create(create_parents)
    for keyval in properties:
        dataset.add_local_property(keyval.key, keyval.val)

    return dataset


def destroy(filesystem, recursive):
    '''Remove a filesystem.'''
    dataset = get_dataset_by(filesystem, should_be=Filesystem)
    dataset.destroy(recursive)
    return dataset


def diff(identifier, other_identifier):
    '''Diff a snapshot against another snapshot in the same filesystem, or
    against the current working filesystem.
    '''
    dataset1 = get_dataset_by(identifier, should_be=Snapshot)
    dataset2 = get_dataset_by(other_identifier)
    # real ZFS can't diff snapshots in different filesystem; not so in ZzzFS
    #if isinstance(dataset2, Filesystem) and (
    #        dataset1.filesystem.name != dataset2.filesystem.name):
    #    raise ZzzFSException(
    #        '%s: cannot compare to a different filesystem' % identifier)

    output = []
    def do_diff(dcmp):
        # trim off pool root from diff output
        base_path = dcmp.left[len(dataset1.data)+1:]
        for name in dcmp.diff_files:
            output.append('M\t%s' % os.path.join(base_path, name))
        for name in dcmp.left_only:
            output.append('-\t%s' % os.path.join(base_path, name))
        for name in dcmp.right_only:
            output.append('+\t%s' % os.path.join(base_path, name))
        for sub_dcmp in dcmp.subdirs.values():
            do_diff(sub_dcmp)
    do_diff(filecmp.dircmp(dataset1.data, dataset2.data))

    return '\n'.join(output)


def get(properties, identifiers, headers, sources, scriptable_mode, recursive,
        max_depth, types):
    '''Get a set of properties for a set of datasets.'''
    all_headers = ['name', 'property', 'value', 'source']
    if headers.items == ['all']:
        headers.items = all_headers
    headers.validate_against(all_headers)
    sources.validate_against(['local', 'inherited'])

    attrs = []
    for dataset in get_all_datasets(identifiers, types, recursive, max_depth):
        if properties.items == ['all']:
            if 'local' in sources.items:
                for key, val in dataset.get_local_properties().items():
                    attrs.append({
                        'name': dataset.name, 'property': key, 'value': val,
                        'source': 'local'})

            if 'inherited' in sources.items:
                for key, val in dataset.get_inherited_properties().items():
                    attrs.append({
                        'name': dataset.name, 'property': key, 'value': val,
                        'source': 'inherited'})

        else:
            for p in properties.items:
                val, source = dataset.get_property_and_source(p)
                if source in sources.items:
                    attrs.append({
                        'name': dataset.name, 'property': p, 'value': val,
                        'source': source})

    return tabulated(attrs, headers, scriptable_mode)


def inherit(property, identifiers):
    '''Remove a local property from a set of datasets.'''
    if not validate_component_name(property):
        raise ZzzFSException('%s: invalid property' % property)

    datasets = [get_dataset_by(identifier) for identifier in identifiers]
    for dataset in datasets:
        try:
            os.remove(os.path.join(dataset.properties, property))
        except OSError:
            # property was not set locally
            pass
    return datasets


def list(identifiers, types, scriptable_mode, headers, recursive, max_depth,
         sort_asc, sort_desc):
    '''Tabulate a set of properties for a set of datasets.'''
    records = []
    for d in get_all_datasets(identifiers, types, recursive, max_depth):
        records.append(dict((h, d.get_property(h)) for h in headers.names))

    return tabulated(records, headers, scriptable_mode, sort_asc, sort_desc)


def promote(clone_filesystem):
    '''Turn a cloned snapshot into a standalone filesystem.'''
    # Since there are no actual dependencies in ZzzFS, this is a no-op.
    dataset = get_dataset_by(
        clone_filesystem, should_be=Filesystem, should_exist=True)
    return dataset


def receive(filesystem, stream=sys.stdin):
    '''Create a new filesystem pre-populated with the contens of a snapshot
    sent via zzzfs send piped through stdin.
    '''
    dataset = get_dataset_by(
        filesystem, should_be=Filesystem, should_exist=False)
    dataset.create(from_stream=stream)
    return dataset


def rename(identifier, other_identifier):
    '''Move or rename the dataset.'''
    dataset1 = get_dataset_by(identifier)
    dataset2 = None  # may be filesystem or snapshot, will check below

    if isinstance(dataset1, Snapshot):
        if not '@' in other_identifier:
            # second argument might be snapshot alone, which we'd interpret as
            # a filesystem; e.g. "rename fs@snapshot new_snapshot"
            other_identifier = '%s@%s' % (
                dataset1.filesystem.name, other_identifier)

        # re-identify with should_exist
        dataset2 = get_dataset_by(
            other_identifier, should_be=Snapshot, should_exist=False)

        # both snapshots
        if dataset1.filesystem.name != dataset2.filesystem.name:
            raise ZzzFSException('mismatched filesystems')

    else:  # dataset1 is a filesystem
        dataset2 = get_dataset_by(
            other_identifier, should_be=Filesystem, should_exist=False)

        if dataset1.pool.name != dataset2.pool.name:
            raise ZzzFSException('cannot rename to different pool')

    # same procedure whether filesystem or snapshot
    dataset1.rename(dataset2)
    return [dataset1, dataset2]


def rollback(snapshot):
    '''Replace the filesystem with the contents of the spceified snapshot.'''
    dataset = get_dataset_by(snapshot, should_be=Snapshot)
    dataset.filesystem.rollback_to(dataset)
    return dataset


def send(snapshot, stream=sys.stdout):
    '''Create a gzipped tarball of a snapshot and write it to sdout.'''
    dataset = get_dataset_by(snapshot, should_be=Snapshot)
    dataset.to_stream(stream)
    return dataset


def set(keyval, identifiers):
    '''Set a property value for a set of datasets.'''
    datasets = [get_dataset_by(identifier) for identifier in identifiers]
    for dataset in datasets:
        dataset.add_local_property(keyval.key, keyval.val)
    return datasets


def snapshot(snapshots, properties):
    '''Create a snapshot of a filesystem.'''
    for i in snapshots:
        dataset = get_dataset_by(i, should_be=Snapshot, should_exist=False)
        if not dataset.filesystem.exists():
            raise ZzzFSException(
                '%s: no such filesystem' % dataset.filesystem.name)
        dataset.create()
        for keyval in properties:
            dataset.add_local_property(keyval.key, keyval.val)
        yield dataset



#!/usr/bin/env python2.7
#
# CDDL HEADER START
#
# The contents of this file are subject to the terms of the
# Common Development and Distribution License, version 1.1 (the "License").
# You may not use this file except in compliance with the License.
#
# You can obtain a copy of the license at ./LICENSE.
# See the License for the specific language governing permissions
# and limitations under the License.
#
# When distributing Covered Code, include this CDDL HEADER in each
# file and include the License file at ./LICENSE.
# If applicable, add the following below this CDDL HEADER, with the
# fields enclosed by brackets "[]" replaced with your own identifying
# information: Portions Copyright [yyyy] [name of copyright owner]
#
# CDDL HEADER END
#

# Copyright (c) 2015 Daniel W. Steinbrook. All rights reserved.

#
# ZzzFS strucutre:
#
#   <ZZZFS_ROOT>/
#     <pool_name>/
#       data -> <disk>
#       properties/
#       filesystems/
#         <fs_name>/
#           data -> ../data/<fs_name>/
#           properties/
#           snapshots/
#             <snapshot_name>/
#               data/
#               properties/
#             [...]
#         <fs_name>%<sub_fs_name>/
#           data -> ../data/<fs_name>/<sub_fs_name>/
#           properties/
#           snapshots/
#         [...]
#     [...]

import io
import os
import csv
import pwd
import gzip
import shutil
import logging
import tarfile
import datetime
import platform

from libzzzfs.util import validate_component_name, ZzzFSException

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)

ZZZFS_DEFAULT_ROOT = os.path.expanduser('~/.zzzfs')


def get_dataset_by(dataset_name, should_be=None, should_exist=True):
    '''Handle user-specified dataset name, returning a Filesystem or Snapshot
    based on the name. If should_be is specified, an exception is raised if the
    dataset is not an instance of the specified class. If should_exist is
    False/True, an exception is raised if the dataset does/does not already
    exist; no check is performed if should_exist is None.
    '''
    # validate dataset identifier
    filesystem_name = dataset_name
    snapshot_name = None
    # distinguish between "fs_name" and "fs_name@snapshot"
    if dataset_name.count('@') == 1:
        filesystem_name, snapshot_name = dataset_name.split('@', 1)

    if not validate_component_name(filesystem_name, allow_slashes=True):
        raise ZzzFSException('%s: invalid dataset identifier' % dataset_name)

    obj = Filesystem(dataset_name)
    if snapshot_name:
        if not validate_component_name(snapshot_name):
            raise ZzzFSException('%s: invalid snapshot name' % snapshot_name)

        obj = Snapshot(filesystem_name, snapshot_name)

    if should_be:
        if not isinstance(obj, should_be):
            raise ZzzFSException(
                '%s: not a %s' % (dataset_name, should_be.__name__.lower()))

    if should_exist and not obj.exists():
        raise ZzzFSException('%s: no such dataset' % dataset_name)
    if obj.exists() and should_exist == False:
        raise ZzzFSException('%s: dataset exists' % dataset_name)

    # pool should exist, even if dataset itself shouldn't
    #logger.debug('%s, in pool %s', obj, obj.pool)
    if not obj.pool.exists():
        raise ZzzFSException('%s: no such pool' % obj.pool.name)

    return obj


def get_all_datasets(identifiers, types, recursive, max_depth):
    '''Get all datasets matching the given identifier names and dataset types,
    and optionally all or a generational subset of their descendants.
    '''
    types.validate_against(['all', 'filesystems', 'snapshots', 'snap'])

    # start with set of all filesystems and snapshots
    filesystems = [f for p in Pool.all() for f in p.get_filesystems()]
    snapshots = [s for f in filesystems for s in f.get_snapshots()]
    datasets = filesystems + snapshots

    # filter to specific identifiers if requested
    if identifiers:
        datasets = [get_dataset_by(i) for i in identifiers]

        # add children of specified identifiers, if requested
        if recursive or max_depth:
            children = []
            for d in datasets:
                if isinstance(d, Filesystem):
                    children += d.get_children(max_depth)
            datasets += children

        # add any snapshots of identifiers and their descendants
        # it's safe to modify the list as we iterate, because we're only adding
        # snapshots, not filesystems
        for d in datasets:
            if isinstance(d, Filesystem):
                datasets += d.get_snapshots()

    # filter out filesystems, if not requested
    if not any(t in ('all', 'filesystems') for t in types.items):
        datasets = [d for d in datasets if not isinstance(d, Filesystem)]

    # filter out snapshots, if not requested
    if not any(t in ('all', 'snapshots', 'snap') for t in types.items):
        datasets = [d for d in datasets if not isinstance(d, Snapshot)]

    return datasets


class Dataset(object):
    '''Base class for Pool, Filesystem, and Snapshot. Contains methods that
    apply to all three objects.
    '''
    def __repr__(self):
        return '<%s: %s>' % (self.__class__.__name__, self.name)

    @property
    def properties(self):
        return os.path.join(self.root, 'properties')

    @property
    def data(self):
        return os.path.join(self.root, 'data')

    @property
    def base_attrs(self):
        return {'name': self.name}

    def get_parent(self):
        if '/' in self.name:
            return Filesystem(self.name.rsplit('/', 1)[-2])
        return Pool(self.name)

    def get_local_properties(self):
        attrs = self.base_attrs
        if not os.path.exists(self.properties):
            # no local attributes
            return attrs

        for key in os.listdir(self.properties):
            with open(os.path.join(self.properties, key), 'r') as f:
                attrs[key] = f.read()

        #logger.debug('%s local attributes: %s', self.name, attrs)
        return attrs

    def get_inherited_properties(self):
        attrs = {}
        local_attrs = self.get_local_properties()

        # inherit values for any attributes not overridden locally, bottom-up
        parent = self
        while parent.get_parent():
            parent = parent.get_parent()
            for key, val in parent.get_local_properties().items():
                if key not in attrs and key not in local_attrs:
                    attrs[key] = val

        return attrs

    def add_local_property(self, key, val):
        if not os.path.exists(self.properties):
            os.makedirs(self.properties)
        if '/' in key:
            raise ZzzFSException('%s: invalid property' % key)
        with open(os.path.join(self.properties, key), 'w') as f:
            f.write(val)

    def get_property_and_source(self, key):
        local = self.get_local_properties()
        if key in local:
            return (local[key], 'local')

        inherited = self.get_inherited_properties()
        if key in inherited:
            return (inherited[key], 'inherited')

        # property not found
        return (None, None)

    def get_property(self, key):
        val, _ = self.get_property_and_source(key)
        return val


class Pool(Dataset):
    def __init__(self, name, should_exist=None):
        self.name = name

        zzzfs_root = os.environ.get('ZZZFS_ROOT', ZZZFS_DEFAULT_ROOT)
        if not os.path.exists(zzzfs_root):
            os.makedirs(zzzfs_root)

        self.root = os.path.join(zzzfs_root, self.name)
        self.filesystems = os.path.join(self.root, 'filesystems')
        self.history = os.path.join(self.root, 'history')

        if should_exist and not self.exists():
            raise ZzzFSException('%s: no such pool' % self.name)
        if should_exist == False and self.exists():
            raise ZzzFSException('%s: pool exists' % self.name)

    def get_parent(self):
        # pool is the top-most desendent of any dataset
        return None

    @classmethod
    def all(self):
        # return an array of all Pool objects
        try:
            return [Pool(name) for name in os.listdir(
                os.environ.get('ZZZFS_ROOT', ZZZFS_DEFAULT_ROOT))]
        except OSError:
            # zzzfs_root doesn't exist, so no pools have been created
            return []

    def exists(self):
        return self.name in os.listdir(
            os.environ.get('ZZZFS_ROOT', ZZZFS_DEFAULT_ROOT))

    def create(self, disk):
        if os.path.exists(disk) and len(os.listdir(disk)) != 0:
            raise ZzzFSException('%s: disk in use' % self.name)

        os.makedirs(self.root)
        pool_target = os.path.join(os.path.abspath(disk), self.name)
        os.makedirs(pool_target)
        os.symlink(pool_target, self.data)

        # create initial root filesystem for this pool
        Filesystem(self.name).create()

    def destroy(self):
        if os.path.exists(os.path.realpath(self.data)):
            shutil.rmtree(os.path.realpath(self.data))
        shutil.rmtree(self.root)

    def get_filesystems(self):
        # unescape slashes when instantiating Filesystem object
        return [Filesystem(x.replace('%', '/'))
            for x in os.listdir(self.filesystems)]

    def get_history(self, long_format=False):
        try:
            with open(self.history, 'r') as f:
                history = csv.reader(f)
                for (date, command, user, host) in history:
                    if long_format:
                        yield '%s %s [user %s on %s]' % (date, command, user, host)
                    else:
                        yield '%s %s' % (date, command)

        except IOError:
            # no logged history
            pass

    def log_history_event(self, argv, date=None, user=None, host=None):
        command = ' '.join(argv)
        if not date:  # default date is now
            date = datetime.datetime.now()
        if not user:  # default user is user executing this script
            user = pwd.getpwuid(os.getuid()).pw_name
        if not host:  # default host is the current platform host
            host = platform.node()

        with open(self.history, 'a') as f:
            history = csv.writer(f)
            history.writerow(
                [date.strftime('%Y-%m-%d.%H:%M:%S'), command, user, host])


class Filesystem(Dataset):
    def __init__(self, filesystem):
        # need to escape slashes to use filesystem name as file name
        self.name = filesystem
        self.safe_name = self.name.replace('/', '%')

        # get pool name by walking up tree
        obj = self
        while obj.get_parent():
            obj = obj.get_parent()
        self.pool = Pool(obj.name)
        self.poolless_name = self.name[len(self.pool.name)+1:]

        self.root = os.path.join(self.pool.root, 'filesystems', self.safe_name)
        self.snapshots = os.path.join(self.root, 'snapshots')

    @property
    def mountpoint(self):
        # before the filesystem is created, the symlink doesn't resolve, so
        # this is a method that recomputes te property whenever it is accessed
        return os.path.realpath(self.data)

    @property
    def base_attrs(self):
        data = super(Filesystem, self).base_attrs
        data['mountpoint'] = self.mountpoint
        return data

    def exists(self):
        return os.path.exists(self.root)

    def get_children(self, max_depth=0):  # 0 = all descendants
        children = [
            f for f in self.pool.get_filesystems()
            if f.name.startswith(self.name + '/')]
        #logger.debug('%s children: %s', self, children)

        if max_depth > 0:
            # use number of slashes to count depth
            depth = max_depth + self.name.count('/')
            children = [f for f in children if f.name.count('/') <= depth]

        return children

    def get_snapshots(self):
        return [Snapshot(self.name, x) for x in os.listdir(self.snapshots)]

    def create(self, create_parents=False, from_stream=None):
        if not self.get_parent().exists():
            if create_parents:
                #logger.debug('%s: need to create %s', self, self.get_parent())
                self.get_parent().create(create_parents=True)
            else:
                raise ZzzFSException(
                    '%s: parent filesystem missing' % self.name)

        # create relative symlink into pool data
        target = os.path.join('..', '..', 'data', self.poolless_name)
        try:
            os.makedirs(os.path.join(self.root, target))
        except OSError:
            # already exists
            pass
        os.symlink(target, self.data)
        os.makedirs(self.properties)
        os.makedirs(self.snapshots)
        #logger.debug('%s: pointed %s at %s', self, self.data, target)

        if from_stream:
            # for receive command: inverse of Snapshot.to_stream
            try:
                # gzip needs a seekable object, not a stream
                #XXX this entails fitting the entire snapshot itno memeory
                buf = io.BytesIO(from_stream.read())
                buf.seek(0)
                with gzip.GzipFile(fileobj=buf) as g:
                    with tarfile.TarFile(fileobj=g) as t:
                        #logger.debug('files in stream: %s', t.getnames())
                        # extract into snapshots directory
                        t.extractall(self.snapshots)

                        # "rollback" filesystem to snapshot just received
                        self.rollback_to(
                            Snapshot(self.name, os.listdir(self.snapshots)[0]))

            except Exception as e:
                # if anything goes wrong, destroy target filesystem and exit
                self.destroy()
                raise ZzzFSException(e)

        #logger.debug(
        #    'after creating %s, filesystems in %s: %s', self, self.pool,
        #    self.pool.get_filesystems())

    def destroy(self, recursive=False):
        dependencies = [
            f for f in self.pool.get_filesystems()
            if f.name.startswith(self.name + '/')]
        #logger.debug('%s dependencies: %s', self, dependencies)

        if len(dependencies) > 0 and not recursive:
            raise ZzzFSException(
                'cannot destroy %r: filesystem has children\n'
                'use \'-r\' to destroy the following datasets:\n'
                '%s' % (self.name, '\n'.join(f.name for f in dependencies)))

        # user may have already deleted data
        if os.path.exists(self.mountpoint):
            shutil.rmtree(self.mountpoint)
        shutil.rmtree(self.root)

        # delete any child filesystems
        for f in dependencies:
            f.destroy(recursive)

    def rollback_to(self, snapshot):
        shutil.rmtree(self.mountpoint)
        shutil.copytree(snapshot.data, self.mountpoint)

        # restore any local properties
        if os.path.exists(snapshot.properties):
            shutil.rmtree(self.properties)
            shutil.copytree(snapshot.properties, self.properties)

    def rename(self, new_dataset):
        # re-create relative symlink into pool data
        target = os.path.join('..', '..', 'data', new_dataset.poolless_name)
        try:
            os.makedirs(os.path.join(new_dataset.root, target))
        except OSError:
            # already exists
            pass

        # move each component individually
        os.symlink(target, new_dataset.data)

        # shutil.move treats destination as parent if it is a directory
        #logger.debug(
        #    '%s: %s -> %s', self, self.mountpoint, new_dataset.mountpoint)
        os.rmdir(new_dataset.mountpoint)
        shutil.move(self.mountpoint, new_dataset.mountpoint)
        shutil.move(self.properties, new_dataset.root)
        shutil.move(self.snapshots, new_dataset.root)

        # all data has been moved
        self.destroy()


class Snapshot(Dataset):
    def __init__(self, filesystem, snapshot):
        self.filesystem = Filesystem(filesystem)
        self.name = snapshot
        self.full_name = '%s@%s' % (filesystem, snapshot)
        self.root = os.path.join(self.filesystem.root, 'snapshots', self.name)
        self.pool = self.filesystem.pool

    @property
    def base_attrs(self):
        data = super(Snapshot, self).base_attrs
        data['name'] = self.full_name
        return data

    def exists(self):
        return os.path.exists(self.root)

    def create(self):
        os.makedirs(self.root)
        shutil.copytree(self.filesystem.data, self.data)
        if os.path.exists(self.filesystem.properties):
            shutil.copytree(self.filesystem.properties, self.properties)
        else:
            # no local properties associated with current working filesystem;
            #  use an empty directory for the snapshot's filesystem
            os.makedirs(self.properties)

    def rename(self, new_snapshot):
        os.rename(self.root, new_snapshot.root)

    def clone_to(self, new_filesystem):
        new_filesystem.create()
        #logger.debug('%s: cloning to %s', self, new_filesystem.mountpoint)

        # remove folders to be replaced by copytree
        #logger.debug(
        #    '%s: %s -> %s', self, self.data, new_filesystem.mountpoint)
        os.rmdir(new_filesystem.mountpoint)
        os.rmdir(new_filesystem.properties)
        shutil.copytree(self.data, new_filesystem.mountpoint)
        shutil.copytree(self.properties, new_filesystem.properties)

    def to_stream(self, stream):
        # write a gzipped tar of the snapshot to the stream
        with gzip.GzipFile(fileobj=stream, mode='w') as g:
            with tarfile.open(fileobj=g, mode='w') as t:
                t.add(self.root, arcname=self.name)



#!/usr/bin/env python2.7
#
# CDDL HEADER START
#
# The contents of this file are subject to the terms of the
# Common Development and Distribution License, version 1.1 (the "License").
# You may not use this file except in compliance with the License.
#
# You can obtain a copy of the license at ./LICENSE.
# See the License for the specific language governing permissions
# and limitations under the License.
#
# When distributing Covered Code, include this CDDL HEADER in each
# file and include the License file at ./LICENSE.
# If applicable, add the following below this CDDL HEADER, with the
# fields enclosed by brackets "[]" replaced with your own identifying
# information: Portions Copyright [yyyy] [name of copyright owner]
#
# CDDL HEADER END
#

# Copyright (c) 2015 Daniel W. Steinbrook. All rights reserved.

from libzzzfs.dataset import Pool
from libzzzfs.util import tabulated, ZzzFSException


def create(pool_name, disk):
    '''Add a pool in the specified directory.'''
    pool = Pool(pool_name, should_exist=False)
    pool.create(disk)
    return pool


def destroy(pool_name):
    '''Remove a pool.'''
    Pool(pool_name, should_exist=True).destroy()


def history(pool_names, long_format):
    pools = Pool.all()
    if pool_names:
        pools = [Pool(p, should_exist=True) for p in pool_names]

    output = []
    for pool in pools:
        # Each pool will have at least one history record (zzzpool create).
        output.append('History for %r:' % pool.name)
        output += pool.get_history(long_format)

    return '\n'.join(output)


def list(pool_name, headers, scriptable_mode):
    '''List all pools.'''
    headers.validate_against([
        'name', 'size', 'alloc', 'free', 'cap', 'health', 'altroot'])

    pools = Pool.all()
    if pool_name:
        pools = [Pool(pool_name, should_exist=True)]

    return tabulated(
        [{'name': p.name, 'health': 'ONLINE'} for p in pools], headers,
        scriptable_mode)



#!/usr/bin/env python2.7
#
# CDDL HEADER START
#
# The contents of this file are subject to the terms of the
# Common Development and Distribution License, version 1.1 (the "License").
# You may not use this file except in compliance with the License.
#
# You can obtain a copy of the license at ./LICENSE.
# See the License for the specific language governing permissions
# and limitations under the License.
#
# When distributing Covered Code, include this CDDL HEADER in each
# file and include the License file at ./LICENSE.
# If applicable, add the following below this CDDL HEADER, with the
# fields enclosed by brackets "[]" replaced with your own identifying
# information: Portions Copyright [yyyy] [name of copyright owner]
#
# CDDL HEADER END
#

# Copyright (c) 2015 Daniel W. Steinbrook. All rights reserved.


def validate_component_name(component_name, allow_slashes=False):
    '''Check that component name starts with an alphanumeric character, and
    disalllow all non-alphanumeric characters except underscore, hyphen, colon,
    and period in component names.
    '''
    allowed = ('_', '-', ':', '.')
    if allow_slashes:
        allowed += ('/',)

    if len(component_name) == 0:
        return False
    if not component_name[0].isalnum():
        return False
    for c in component_name:
        if c not in allowed and not c.isalnum():
            return False

    return True


class ZzzFSException(Exception):
    pass


class PropertyList(object):
    # Numeric columns are right-aligned when tabulated.
    numeric_types = ['alloc', 'avail', 'cap', 'free', 'refer', 'size', 'used']

    # synonymous field names
    shorthand = {'available': 'avail', 'capacity': 'cap'}

    def __repr__(self):
        return '<%s: %s>' % (self.__class__.__name__, ','.join(self.items))

    def __init__(self, user_string):
        self.items = user_string.split(',')

    def validate_against(self, acceptable):
        # compare against a set of acceptable fields
        for col in self.names:
            if col not in acceptable:
                raise ZzzFSException('%s: unrecognized property name' % col)

    @property
    def names(self):
        # use shorthand name, if any, as canonical name
        for col in self.items:
            yield self.shorthand.get(col, col)

    @property
    def types(self):
        # strings unless explicitly numeric
        for col in self.names:
            if col in self.numeric_types:
                yield int
            else:
                yield str


class PropertyAssignment(object):
    '''property=value command-line argument, as used in zzzfs set command.'''
    def __init__(self, user_string):
        try:
            self.key, self.val = user_string.split('=')
        except ValueError:
            raise ZzzFSException(
                '%r: invalid property=value format' % user_string)

        if not validate_component_name(self.key):
            raise ZzzFSException('%s: invalid property' % self.key)


def tabulated(data, headers, scriptable_mode=False, sort_asc=[], sort_desc=[]):
    '''Generates a printable table as a string given data (an array of dicts)
    and an array of field names for headers.
    '''
    if len(data) == 0:
        return ''

    types = list(headers.types)
    names = list(headers.names)

    row_format = '\t'.join('%s' for i in range(len(names)))
    if not scriptable_mode:
        # For evenly-spaced columns, left-align each text field (right-align
        # each numeric field) in a cell that's big enough for the longest value
        # in each column.
        data_and_headers = data + [dict(zip(names, names))]
        cells = []
        for i in range(len(names)):
            box_width = max(len(r.get(names[i]) or '-') for r in data_and_headers)
            if types[i] == str:
                box_width *= -1  # negative field width means left-align
            cells.append('%%%ds' % box_width)
        row_format = '\t'.join(cells)

    # sort by specified fields, if any
    for field in sort_asc + sort_desc:
        if field not in names:
            raise ZzzFSException('%s: no such column' % field)

    for field in sort_asc:
        data = sorted(data, key=lambda row: row[field])

    for field in sort_desc:
        data = list(reversed(sorted(data, key=lambda row: row[field])))

    # Add individual data rows.
    output = '\n'.join(
        row_format % tuple(row.get(names[i]) or '-' for i in range(len(names)))
        for row in data)

    # Prepend header row in all caps.
    if not scriptable_mode:
        output = row_format % tuple(h.upper() for h in names) + '\n' + output

    return output



# encoding=utf-8
from setuptools import setup

setup(
    name='zzzfs',
    version='0.1',
    description='Dataset management à la ZFS',
    long_description=open('README').read(),
    author='Daniel W. Steinbrook',
    author_email='steinbro@post.harvard.edu',
    url='https://github.com/steinbro/zzzfs',
    license="CDDL",
    keywords='zfs',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Topic :: System :: Filesystems',
    ],
    packages=['libzzzfs', 'libzzzfs.cmd'],
	test_suite='tests',
    entry_points={
        'console_scripts': [
            'zzzfs = libzzzfs.cmd.zzzfs:main',
            'zzzpool = libzzzfs.cmd.zzzpool:main',
        ],
    },
)
